@@ -1,7 +1,7 @@
 use core::fmt::Debug;
 use thiserror::Error;
 
-use crate::{Generation, UniqueSectionId};
+use crate::Generation;
 
 #[derive(Error, Debug)]
 pub enum Error {
@@ -20,16 +20,10 @@ pub enum Error {
         child_gen: Generation,
         merge_gen: Generation,
     },
-    #[error("A vote is always for the next generation: vote gen {vote_gen} != {gen} + 1, pending gen: {pending_gen}")]
-    VoteNotForNextGeneration {
+    #[error("A vote must be tagged with a generation between 1..=(gen={gen} + 1): vote gen = {vote_gen}, gen: {gen}")]
+    VoteForBadGeneration {
         vote_gen: Generation,
         gen: Generation,
-        pending_gen: Generation,
-    },
-    #[error("Vote received has a different unique section id: vote gen {vote_gen} != {gen}")]
-    VoteWithInvalidUniqueSectionId {
-        vote_gen: UniqueSectionId,
-        gen: UniqueSectionId,
     },
     #[error("The voter is not an elder")]
     NotElder,
