@@ -1,33 +1,23 @@
 use std::collections::{BTreeMap, BTreeSet};
 
-use blsttc::{PublicKeySet, SecretKeyShare, Signature};
+use blsttc::{PublicKeySet, SecretKeyShare};
 use core::fmt::Debug;
 use log::info;
 use serde::{Deserialize, Serialize};
 
 use crate::consensus::{Consensus, VoteResponse};
 use crate::vote::{Ballot, Proposition, SignedVote, Vote};
-use crate::{Decision, Error, Fault, NodeId, Result};
+use crate::{Error, NodeId, Result};
 
 const SOFT_MAX_MEMBERS: usize = 7;
 pub type Generation = u64;
 
-#[derive(Debug)]
-pub struct HistoryEntry<T: Proposition> {
-    pub votes: BTreeSet<SignedVote<Reconfig<T>>>,
-    pub proposals: BTreeMap<Reconfig<T>, Signature>,
-    pub faults: BTreeSet<Fault<Reconfig<T>>>,
-}
-
 #[derive(Debug)]
 pub struct Membership<T: Proposition> {
     pub consensus: Consensus<Reconfig<T>>,
-    // TODO: we need faulty elder detection
-    // pub faulty_elders: BTreeMap<PublicKeyShare, BTreeSet<SignedVote<Reconfig<T>>>>,
     pub gen: Generation,
-    pub pending_gen: Generation,
-    pub forced_reconfigs: BTreeMap<Generation, BTreeSet<Reconfig<T>>>, // TODO: change to bootstrap members
-    pub history: BTreeMap<Generation, HistoryEntry<T>>, // for onboarding new procs, the vote proving super majority
+    pub forced_reconfigs: BTreeMap<Generation, BTreeSet<Reconfig<T>>>,
+    pub history: BTreeMap<Generation, Consensus<Reconfig<T>>>,
 }
 
 #[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Serialize, Deserialize)]
@@ -60,12 +50,27 @@ impl<T: Proposition> Membership<T> {
         elders: PublicKeySet,
         n_elders: usize,
     ) -> Self {
-        Membership::<T> {
-            consensus: Consensus::<Reconfig<T>>::from(secret_key, elders, n_elders),
+        Membership {
+            consensus: Consensus::from(secret_key, elders, n_elders),
             gen: 0,
-            pending_gen: 0,
             forced_reconfigs: Default::default(),
-            history: BTreeMap::new(),
+            history: BTreeMap::default(),
+        }
+    }
+
+    pub fn consensus_at_gen(&self, gen: Generation) -> Option<&Consensus<Reconfig<T>>> {
+        if gen == self.gen + 1 {
+            Some(&self.consensus)
+        } else {
+            self.history.get(&gen)
+        }
+    }
+
+    pub fn consensus_at_gen_mut(&mut self, gen: Generation) -> Option<&mut Consensus<Reconfig<T>>> {
+        if gen == self.gen + 1 {
+            Some(&mut self.consensus)
+        } else {
+            self.history.get_mut(&gen)
         }
     }
 
@@ -99,7 +104,7 @@ impl<T: Proposition> Membership<T> {
             return Ok(members);
         }
 
-        for (history_gen, history_entry) in self.history.iter() {
+        for (history_gen, consensus) in self.history.iter() {
             self.forced_reconfigs
                 .get(history_gen)
                 .cloned()
@@ -107,7 +112,16 @@ impl<T: Proposition> Membership<T> {
                 .into_iter()
                 .for_each(|r| r.apply(&mut members));
 
-            for (reconfig, _sig) in history_entry.proposals.iter() {
+            let decision = if let Some(decision) = consensus.decision.as_ref() {
+                decision
+            } else {
+                panic!(
+                    "historical consensus entry without decision {}: {:?}",
+                    history_gen, consensus
+                );
+            };
+
+            for (reconfig, _sig) in decision.proposals.iter() {
                 reconfig.apply(&mut members);
             }
 
@@ -130,7 +144,7 @@ impl<T: Proposition> Membership<T> {
         self.consensus
             .detect_byzantine_voters(&signed_vote)
             .map_err(|_| Error::AttemptedFaultyProposal)?;
-        Ok(self.cast_vote(signed_vote))
+        self.cast_vote(signed_vote)
     }
 
     pub fn anti_entropy(&self, from_gen: Generation) -> Result<Vec<SignedVote<Reconfig<T>>>> {
@@ -140,9 +154,9 @@ impl<T: Proposition> Membership<T> {
             .history
             .iter() // history is a BTreeSet, .iter() is ordered by generation
             .filter(|(gen, _)| **gen > from_gen)
-            .map(|(gen, history_entry)| {
-                self.consensus
-                    .build_super_majority_vote(history_entry.votes.clone(), *gen)
+            .filter_map(|(gen, c)| c.decision.clone().map(|d| (gen, c, d)))
+            .map(|(gen, c, decision)| {
+                c.build_super_majority_vote(decision.votes, decision.faults, *gen)
             })
             .collect::<Result<Vec<_>>>()?;
 
@@ -161,34 +175,22 @@ impl<T: Proposition> Membership<T> {
         signed_vote: SignedVote<Reconfig<T>>,
     ) -> Result<VoteResponse<Reconfig<T>>> {
         self.validate_signed_vote(&signed_vote)?;
-        self.log_signed_vote(&signed_vote);
 
-        let vote_response = self
-            .consensus
-            .handle_signed_vote(signed_vote, self.pending_gen)?;
+        let vote_gen = signed_vote.vote.gen;
 
-        match &vote_response {
-            VoteResponse::Broadcast(vote) => {
-                self.pending_gen = vote.vote.gen;
-            }
-            VoteResponse::Decided(Decision {
-                votes,
-                proposals,
-                faults,
-            }) => {
-                self.history.insert(
-                    self.pending_gen,
-                    HistoryEntry {
-                        votes: votes.clone(),
-                        proposals: proposals.clone(),
-                        faults: faults.clone(),
-                    },
-                );
-                self.gen = self.pending_gen;
-                // clear our pending votes
-                self.consensus.votes = Default::default();
-            }
-            VoteResponse::WaitingForMoreVotes => {}
+        let consensus = self.consensus_at_gen_mut(vote_gen).unwrap();
+        let vote_response = consensus.handle_signed_vote(signed_vote)?;
+
+        if consensus.decision.is_some() && vote_gen == self.gen + 1 {
+            let next_consensus = Consensus::from(
+                self.consensus.secret_key.clone(),
+                self.consensus.elders.clone(),
+                self.consensus.n_elders,
+            );
+
+            let decided_consensus = std::mem::replace(&mut self.consensus, next_consensus);
+            self.history.insert(vote_gen, decided_consensus);
+            self.gen = vote_gen
         }
 
         Ok(vote_response)
@@ -198,42 +200,33 @@ impl<T: Proposition> Membership<T> {
         self.consensus.sign_vote(vote)
     }
 
-    pub fn cast_vote(&mut self, signed_vote: SignedVote<Reconfig<T>>) -> SignedVote<Reconfig<T>> {
-        self.log_signed_vote(&signed_vote);
-        signed_vote
-    }
-
-    fn log_signed_vote(&mut self, signed_vote: &SignedVote<Reconfig<T>>) {
-        self.pending_gen = signed_vote.vote.gen;
-        self.consensus.log_signed_vote(signed_vote);
-    }
-
-    pub fn count_votes(
-        &self,
-        votes: &BTreeSet<SignedVote<Reconfig<T>>>,
-    ) -> BTreeMap<BTreeSet<Reconfig<T>>, usize> {
-        self.consensus.count_votes(votes)
+    pub fn cast_vote(
+        &mut self,
+        signed_vote: SignedVote<Reconfig<T>>,
+    ) -> Result<SignedVote<Reconfig<T>>> {
+        self.consensus.cast_vote(&signed_vote)?;
+        Ok(signed_vote)
     }
 
     pub fn validate_signed_vote(&self, signed_vote: &SignedVote<Reconfig<T>>) -> Result<()> {
-        if signed_vote.vote.gen != self.gen + 1 {
-            return Err(Error::VoteNotForNextGeneration {
+        if let Some(c) = self.consensus_at_gen(signed_vote.vote.gen) {
+            c.validate_signed_vote(signed_vote)?;
+        } else {
+            return Err(Error::VoteForBadGeneration {
                 vote_gen: signed_vote.vote.gen,
                 gen: self.gen,
-                pending_gen: self.pending_gen,
             });
         }
 
         signed_vote
             .proposals()
             .into_iter()
-            .try_for_each(|reconfig| self.validate_reconfig(reconfig))?;
-
-        self.consensus.validate_signed_vote(signed_vote)
+            .try_for_each(|reconfig| self.validate_reconfig(reconfig, signed_vote.vote.gen))
     }
 
-    pub fn validate_reconfig(&self, reconfig: Reconfig<T>) -> Result<()> {
-        let members = self.members(self.gen)?;
+    pub fn validate_reconfig(&self, reconfig: Reconfig<T>, gen: Generation) -> Result<()> {
+        assert!(gen > 0);
+        let members = self.members(gen - 1)?;
         match reconfig {
             Reconfig::Join(actor) => {
                 if members.contains(&actor) {
